// CSetup.cpp: implementation of the CSetup class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <winsvc.h>
#include "dowset.h"
#include "Csetup.h"
#include "dow95.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSetup::CSetup()
{

}

CSetup::~CSetup()
{

}

//////////////////////////////////////////////////////////////////////
// Install -- Installs the drivers for the correct OS.				//
//		arguments -- none.											//
//		returns -- integer 1-6 or 13 based on instalation result.	//
//////////////////////////////////////////////////////////////////////
int CSetup::Install()
{
	int result;
	OSVERSIONINFO osvi;
	char SystemDir[360];

	memset(&osvi, 0, sizeof(OSVERSIONINFO));
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&osvi);

	if(osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)	// Windows 95 detected.
	{
		if(GetSystemDirectory(SystemDir, 360))
		{
			strcat(SystemDir, "\\vsauthd.vxd");
			if(CopyFile("vsauthd.vxd", SystemDir, false)) // Try to copy the driver.
			{
				if(dowcheck())
				{
			        result = 1;  //Windows 95 Device Driver Installed Successfully.
                    if (this->debugMode)
                    {
                        printf("result = 1 - Win95 driver install successful.\n");
                    }
                }
                else
                {
                    result = 2;  //Failed to install Windows 95 Device Driver.
                    if (this->debugMode)
                    {
                        printf("result = 2 - Win95 driver install NOT successful.\n");
                    }
                }
                
            }
            else if (dowcheck ())
            {
                result = 1;  //Windows 95 Device Driver Installed Successfully.
                if (this->debugMode)
                {
                    printf("result = 1 - Win95 driver install successful.\n");
                }
            }
            else
            {
                result = 2;  //Failed to install Windows 95 Device Driver.
                if (this->debugMode)
                {
                    printf("result = 2 - Win95 driver install NOT successful.\n");
                }
            }
        }
        else
        {
            result = 2;      //Failed to install Windows 95 Device Driver.
            if (this->debugMode)
            {
                printf("result = 2 - Win95 driver install NOT successful.\n");
            }
        }
        
    }
    else if (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT) // Win NT detected.
    {
        if (ThirtyTwoBit)
        {
            if (GetSystemDirectory (SystemDir, 360))
            {
                strcat (SystemDir, "\\drivers\\ds1410d.sys");
                Remove ();
                if (CopyFile ("ds1410d.sys", SystemDir, FALSE)) // Try to copy the physical device driver.
                {
                    if (dowcheck ())
                    {
                        result = 3; //Windows NT Device Driver Installed Successfully.
                        if (this->debugMode)
                        {
                            printf("result = 3 - Both WinNT device drivers installed.\n");
                        }
                        if (SixteenBit)
                        {
                            result = InstallVDD(); // Try to install the virtual device driver.
                        }
                    }
                    else
                    {
                        result = 4; //Failed to install Windows NT Device Driver
                        if (this->debugMode)
                        {
                            printf("result = 4 - WinNT driver install FAILED.\n");
                        }
                    }
                }
                else if (dowcheck ())
                {
                    result = 3; //Windows NT Device Driver Installed Successfully.
                    if (this->debugMode)
                    {
                        printf("result = 3 - Both WinNT device drivers installed.\n");
                    }
                    if (SixteenBit)
                    {
                        result = InstallVDD (); // Try to install the virtual device driver.
                    }
                }
                else
                {
                    result = 4; //Failed to install Windows NT Device Driver.
                    if (this->debugMode)
                    {
                        printf("result = 4 - WinNT driver install FAILED.\n");
                    }
                }
            }
            else
            {
                result = 4; //Failed to install Windows NT Device Driver.
                if (this->debugMode)
                {
                    printf("result = 4 - WinNT driver install FAILED.\n");
                }
            }
            
            // Now check for service pack 2.
            if (!NTVDMDateOK())
            {
                result = 5; //NT 4.0 will require service pack #2 or greater.                                    "in order for these drivers to function correctly.\n"
                if (this->debugMode)
                {
                    printf("result = 5 - WinNT needs to be upgraded to Service Pack #2 or greater.\n");
                }
            }
        }
        
    }
    else
    {
        result = 6; //No Device Driver Needed.
        if (this->debugMode)
        {
            printf("result = 6 - OS not compliant with 95 or NT.\n");
        }
    }

	return result;
    
}

//////////////////////////////////////////////////////////////////////////////
// Remove -- removes device drivers for the correct OS.						//
//		arguments -- none													//
//		returns -- integer 7-12 or 14 based on the result of the removal	//
//////////////////////////////////////////////////////////////////////////////
int CSetup::Remove()
{
    SC_HANDLE hSCManager, hDS1410D;
    BOOL Ret;
    char Service[] = "DS1410D";
    OSVERSIONINFO osvi;
    char SystemDir[360];
	int result;
    
    memset (&osvi, 0, sizeof (OSVERSIONINFO));
    osvi.dwOSVersionInfoSize = sizeof  (OSVERSIONINFO);
    GetVersionEx (&osvi);

    if (osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) // Win 95 detected
    {
        if (GetSystemDirectory (SystemDir, 360))
        {
            strcat (SystemDir,"\\vsauthd.vxd");
            if (DeleteFile (SystemDir))	// Try to delete the driver.
            {
                result = 7; //Windows 95 Device Driver Removed Successfully.
                if (this->debugMode)
                {
                    printf("result = 7 - Win95 device driver removal Successful.\n");
                }
            }
            else
            {
                result = 8; //Failed to delete Windows 95 Device Driver.
							//Make sure no other programs are currently using the driver.
                if (this->debugMode)
                {
                    printf("result = 8 - Win95 device driver removal FAILED.\n");
                }
            }
        }
        else
        {
            result = 9; //Failed to delete Windows 95 Device Driver.");
            if (this->debugMode)
            {
                printf("result = 9 - Win95 device driver removal FAILED (system dir problem).\n");
            }
        }
        
    }
    else if (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT) // Win NT detected.
    {
        if (ThirtyTwoBit)
        {
            SERVICE_STATUS ServiceStatus;      
            /* Get a handle to the Service control manager database */
            hSCManager = OpenSCManager (
                NULL,
                NULL,
                SC_MANAGER_ALL_ACCESS);
            if (hSCManager)
            {
                hDS1410D = OpenService (
                    hSCManager,	
                    Service,	
                    SERVICE_ALL_ACCESS);
                if (hDS1410D)
                {
                    Ret = ControlService (hDS1410D, // handle to service 
                        SERVICE_CONTROL_STOP,	    // control code 
                        &ServiceStatus 	            // pointer to service status structure  
                        );
                    if (DeleteService (hDS1410D))
                    {
                        result = 10; //Successfully deleted Windows NT device driver.
                        GetSystemDirectory (SystemDir, 360);
                        strcat (SystemDir,"\\drivers\\ds1410d.sys");
                        DeleteFile (SystemDir);
                        if (this->debugMode)
                        {
                            printf("result = 10 - WinNT device driver removal Successful.\n");
                        }
                    }
                    else if (GetLastError () == ERROR_SERVICE_MARKED_FOR_DELETE)  
                    // Driver will be deleted when all handles have been closed.
                    {
                        result = 10; //Successfully deleted Windows NT device driver.
                        GetSystemDirectory (SystemDir, 360);
                        strcat (SystemDir,"\\drivers\\ds1410d.sys");
                        DeleteFile (SystemDir);
                        if (this->debugMode)
                        {
                            printf("result = 10 - WinNT device driver removal Successful.\n");
                        }
                    }
                    else
                    {
                        result = 11; //Failed to remove Windows NT device driver.
                    }
                    if (this->debugMode)
                    {
                        printf("result = 11 - WinNT device driver removal FAILED.\n");
                    }
                    CloseServiceHandle (hDS1410D);
                }
                else
                {
                    result = 11; //Failed to remove Windows NT device driver.
                    if (this->debugMode)
                    {
                        printf("result = 11 - WinNT device driver removal FAILED.\n");
                    }
                }
                CloseServiceHandle (hSCManager);
            }
            else
            {
                result = 11; //Failed to remove Windows NT device driver.
                if (this->debugMode)
                {
                    printf("result = 11 - WinNT device driver removal FAILED.\n");
                }
            }
        }
        result = RemoveVDD ();
        result = RemoveVDD (); // for some reason that is beyond me, sometimes it takes two tries.
    }
    else
    {
        result = 12; //No Device Driver to delete.
        if (this->debugMode)
        {
            printf("result = 12 - duh, no device driver to delete!\n");
        }
    }

	return(result);
    
} // -- CSetup::Remove -- // 


//////////////////////////////////////////////////////////
// InstallVDD -- Installs the virtual device driver		//
//		arguments -- none								//
//		returns -- integer 3 if successful 13 if not.	//
//////////////////////////////////////////////////////////
int CSetup::InstallVDD()
{ // -- CSetup::InstallVDD -- // 

    char SubKey[] = "System\\CurrentControlSet\\Control\\VirtualDeviceDrivers"; 
    char ValueName[] = "VDD";
    char ValueData[360],ValueD[360];
    char CurrentPath[360], *p;
    DWORD cbData, Type, Size;
    DWORD ErrorValue;
	int result;
    
    HKEY Result;
    GetCurrentDirectory (360, CurrentPath);
    if (GetSystemDirectory (ValueData, 360))
    {
        strcat (CurrentPath, "\\downtvdd.dll");
        strcat (ValueData, "\\downtvdd.dll");
        
        CopyFile (CurrentPath, ValueData, FALSE);
        if ((ErrorValue = RegOpenKeyEx (
            HKEY_LOCAL_MACHINE,	// handle of open key 
            SubKey,	            // address of name of subkey to open 
            0,	                // reserved 
            KEY_ALL_ACCESS,	    // security access mask 
            & Result            // address of handle of open key 
            )) == ERROR_SUCCESS)
        { // -- if RegOpenKeyEx succeeds -- // 
         		 
            cbData = strlen (ValueData);
            RegQueryValueEx (Result,
                ValueName,
                NULL,
                & Type,
                (unsigned char *) ValueD,
                & Size);
            // don't destroy any values that are already there
            if (Size == 2)
            {
                Size--;
            }

            p = ValueD+Size-1;
            strcpy (p, ValueData);
            strcat (p, "\0\0");
            cbData = Size + strlen (ValueData)+1;
            p = ValueD + cbData-1;
            *p = 0;

            if (RegSetValueEx (
                Result,	                // handle of key to set value for  
                ValueName,	            // address of value to set 
                0,	                    // reserved 
                REG_MULTI_SZ,	        // flag for value type 
                (const unsigned char *) ValueD, // address of value data 
                cbData 	                // size of value data 
                ) == ERROR_SUCCESS)
            { // -- RegSetValueEx successful -- // 

                result = 3; //Windows NT device drivers installed successfully.
                if (this->debugMode)
                {
                    printf("result = 3 - Both WinNT device drivers installed.\n");
                }

            } // -- RegSetValueEx successful -- // 
            else
            {
                result = 13; //Physical Device Driver installed correctly.
							 //Virtual Device Driver failed to install.
                if (this->debugMode)
                {
                    printf("result = 13 - only one WinNT device driver installed.\n");
                }
            }

        } // -- if RegOpenKeyEx succeeds -- // 
        else
        {
            result = 13; //Physical Device Driver installed correctly.
                         //Virtual Device Driver failed to install.
            if (this->debugMode)
            {
                printf("result = 13 - only one WinNT device driver installed.\n");
            }
        }
    }
    else
    {
        result = 13; //Physical Device Driver installed correctly.
                     //Virtual Device Driver failed to install.
        if (this->debugMode)
        {
            printf("result = 13 - only one WinNT device driver installed.\n");
        }
    }

	return result;    
    
} // -- CSetupDlg::InstallVDD -- // 

//////////////////////////////////////////////////////////////
// RemoveVDD -- removes the virtual device driver.			//
//		arguments -- none									//
//		returns -- integer: 10 if successful, 14 if not.	//
//////////////////////////////////////////////////////////////
int CSetup::RemoveVDD ()
{ // -- CSetupDlg::RemoveVDD -- // 

    char SubKey[] = "System\\CurrentControlSet\\Control\\VirtualDeviceDrivers"; 
    char ValueName[] = "VDD", *p;
    char ValueData[360], ValueD[360],ValueTemp[20][360],ValueOld[360];
    DWORD cbData, Type, Size, i,j;
    DWORD ErrorValue;
    int len = 0, result;
    
    HKEY Result;
    if (GetSystemDirectory (ValueData, sizeof (ValueOld)))
    {
        strcat (ValueData, "\\downtvdd.dll");
        
        // Using DeleteFile to delete a system driver will not be 
        // successful under Windows NT.  In that case, rename the 
        // file and mark for deletion on reboot.
        if (!DeleteFile (ValueData))
        {
            GetSystemDirectory (ValueOld, sizeof (ValueOld));
            strcat (ValueOld, "\\downtvdd.del");
            
            // This should rename the file.
            MoveFileEx (
                ValueData,	                // address of name of the existing file  
                ValueOld,               	// address of new name for the file 
                MOVEFILE_REPLACE_EXISTING 	// flag to determine how to move file 
                );
            
            // This marks the file for deletion.
            MoveFileEx (
                ValueOld,
                NULL,
                MOVEFILE_DELAY_UNTIL_REBOOT
                );
        }


        RegFlushKey (HKEY_LOCAL_MACHINE);

        if ((ErrorValue = RegOpenKeyEx (
            HKEY_LOCAL_MACHINE,	    // handle of open key 
            SubKey,	                // address of name of subkey to open 
            0,	                    // reserved 
            KEY_ALL_ACCESS,	        // security access mask 
            &Result 	            // address of handle of open key 
            )) == ERROR_SUCCESS)
        { // -- if RegOpenKeyEx succeeds -- // 
            if (RegQueryValueEx (Result,
                ValueName,
                NULL,
                &Type,
                (unsigned char *) ValueD,
                &Size) == ERROR_SUCCESS)
            { // -- if RegQueryValueEx succeeds -- // 
                p = ValueD;
                i = 0;
                // *(p-1) is questionable on the first loop;
                while (*p || *(p-1))
                {
                    strcpy (ValueTemp[i],p);
                    p = p + strlen (ValueTemp[i])+1;
                    i++;
                }
                strcpy (ValueData, "");

                // Build a new string to set in the registry and 
                // remove any references to our string.
                p = ValueData;

                for (j = 0; j < i; j++)
                { // -- for -- // 
                    strupr (ValueTemp[j]);
                    if (!strstr (ValueTemp[j], "DOWNTVDD.DLL"))
                    {
                        strcpy (p, ValueTemp[j]);
                        len = len + strlen (ValueTemp[j])+1;
                        p += strlen (ValueTemp[j])+1;
                    }
                } // -- for -- // 

                *p = 0;
                cbData = len+1;
                if (cbData == 1)
                {
                    *(p+1) = 0;
                    cbData++;
                }
                if (RegSetValueEx (
                    Result,	        // handle of key to set value for  
                    ValueName,	    // address of value to set 
                    0,	            // reserved 
                    REG_MULTI_SZ,	// flag for value type 
                    (const unsigned char *)ValueData,   // address of value data 
                    cbData 	        // size of value data 
                    ) == ERROR_SUCCESS)
                { // -- if RegSetValueEx succeeds -- // 

                    result = 10; //Windows NT device drivers removed successfully.

                } // -- if RegSetValueEx succeeds -- // 
                else
                {
                    result = 14; //Physical Device Driver removed correctly.
                                 //Virtual Device Driver failed to uninstall.
                }
            } // -- if RegQueryValueEx succeeds -- // 
            else
            {
                result = 14; //Physical Device Driver removed correctly.
                             //Please try again to remove Virtual Device Driver.
            }

        } // -- if RegOpenKeyEx succeeds -- // 
        else
        {
            result = 14; //Physical Device Driver removed correctly.
                         //Virtual Device Driver failed to uninstall.
        }
        RegCloseKey (Result);
    }
    else
    {
        result = 14; //Physical Device Driver removed correctly.
                     //Virtual Device Driver failed to uninstall.
    }

	return(result);

} // -- CSetup::RemoveVDD -- // 

//////////////////////////////////////////////////////////////
// NTVDMDateOK -- determines if Win NT has SP2 or greater.	//
//		arguments -- none									//
//		returns -- integer:  0 if under SP2, 1-4 otherwise.	//
//////////////////////////////////////////////////////////////
int CSetup::NTVDMDateOK (void)
{ // CSetup::NTVDMDateOK

    char l_lpsWinDir[360];
    HANDLE l_tempHandle;
    FILETIME l_FileCreationStamp  = { 0 };
    FILETIME l_FileAccessStamp    = { 0 };
    FILETIME l_FileWriteStamp     = { 0 };
    LARGE_INTEGER l_tempInteger    = { 0 };


    // First of all, check version.  If not in Windows NT 4.0, then 
    // Service Pack #2 is not an issue.
    if ((_winmajor != 4) || (_winminor != 0))
    {
        return (1);
    }
    GetWindowsDirectory (l_lpsWinDir, 360);


    // If OS doesn't know its own directory, something is unusual.
    if (l_lpsWinDir == 0)
    {
        return (2);
        // By default,assume human understands what is happening.
    }
    
    strcat (l_lpsWinDir, "\\system32\\NTVDM.EXE");

    // Check for file existance.
    if ((l_tempHandle = CreateFile (l_lpsWinDir, 
            GENERIC_READ,           // Read.
            FILE_SHARE_READ,        // Allow subsequent share access.
            NULL,                   // No priveleges to child proc.
            OPEN_EXISTING,          // Do not create one.
            FILE_ATTRIBUTE_ARCHIVE, // This is expected.
            NULL                    // No file template used.
            )) == INVALID_HANDLE_VALUE)  // If invalid handle, function failed.
    {
        return (3);
        // By default,assume human understands what is happening.
    }


    // Get time stamps.
    if (!GetFileTime (l_tempHandle, &l_FileCreationStamp, &l_FileAccessStamp, &l_FileWriteStamp))
    {
        return (4);
        // By default,assume human understands what is happening.
    }

    CloseHandle (l_tempHandle);

    // MSDN library reccomends assigning timestamps to a LARGE_INTEGER & then performing ops.
    l_tempInteger.LowPart  = (l_FileWriteStamp.dwLowDateTime);
    l_tempInteger.HighPart = (l_FileWriteStamp.dwHighDateTime);

    if (l_tempInteger.QuadPart >= 124949088000000000)   
        // Compares to one second after midnight, 12/13/96
        // 12/14/96 is the date of Service Pack 2 (a little extra room).
    {

        return (1);
    }
    else
    {
        return (0);
    }

} // CSetup::NTVDMDateOK
